module SIG 

(* data is an abstract datatype for almost everything *)
assume type data

type vk = data
type sk = data
type tag = data

assume val sk2vk  : sk -> Tot vk

(* keys are equipped with predicates *)
opaque logic type verified : vk -> data -> Type
type vkey (p:(data -> Type)) = k:vk{verified k == p}
type skey (p:(data -> Type)) = k:sk{verified (sk2vk k) == p}

assume val keygen: p:(data -> Type) -> skey p   
assume val sign:   p:(data -> Type) -> skey p -> d:data{p d} -> tag
assume val verify: p:(data -> Type) -> v:vkey p -> d:data -> tag -> Tot (b:bool{b ==> p d})


module Cert
open SIG
                                
type text = data
type user = data
type cert = data
type cert_chain = data

(* abstract formatting and parsing functions *)
assume val format : list data -> Tot data
assume val parse : d:data -> Tot (s : list data {format s = d})

(* predicate specifying that a user willingly sent some data *)
opaque logic type sent: user -> data -> Type 

(* predicate specifying the properties of a certification key *)
opaque logic type certified : d:data -> Type
assume Certified:
    ((forall u k. {:pattern (format [u;k])}
            certified (format [u;k]) ==> verified k == sent u)
        /\
    (forall k. {:pattern (format [k])}
            certified (format [k]) ==> verified k == certified ))

(* sender and admin functions *)
type sender (u:user) = d:data{sent u d} -> data
type admin = u:user -> (t:text {sent u t} -> data)

(* a simple sending function for signed mail *)
val send: u:user -> skey (sent u) -> cert_chain -> t:text { sent u t } -> 
          data
let send (u:user) (sk:skey (sent u)) (certs:cert_chain) (t:text { sent u t }) =
  let mtag = sign #(sent u) sk t in
  format [t; mtag; certs]


val register : skey certified -> list cert -> u:user -> sender u
let register k l newuser =
 let rootkey = (keygen (sent newuser)) in
 let verkey = (sk2vk rootkey) in
 let ctxt = format [newuser; verkey] in
 assume (certified ctxt);
  send 
    newuser 
    rootkey
    (format(List.append l [ctxt;(sign certified k ctxt)]))
    
  
val extend : list cert -> skey certified -> admin
let extend l k =
 let rootkey = (keygen certified) in 
 let verkey = (sk2vk rootkey) in
 let ctxt = format[verkey] in
 assume (certified ctxt); 
 let ctag = (sign certified k ctxt) in
  register rootkey 
     (List.append l [ctxt;ctag])
 

type verified_mail = 
    | ReceivedFrom: u:user -> d:data { sent u d } -> verified_mail

(* In this example the certificate chain is of fixed length 1 *)
val checkmail: vkey certified -> data -> option (verified_mail)
let checkmail vk0 inbox = 
  (match parse inbox with 
  | [ text; cert; mtag ] ->
     (match parse cert with
      | [ ctxt; ctag ] ->
        (match parse ctxt with
        | [ sender; vk ] ->
          if verify #certified vk0 ctxt ctag then
             if verify #(sent sender) vk text mtag then
               Some (ReceivedFrom sender text)
             else None
          else None
        | _ -> None)
      | _ -> None)
  | _ -> None) 


val check_certificates: vkey certified -> list data -> d:option (data * data) {is_Some d ==> certified (fst (Some.v d))}
let rec check_certificates trust_anchor l = match l with
  | txt::tag::tl -> if verify #certified trust_anchor txt tag 
                   then( match (parse txt) with
                   | [new_trust_anchor] -> check_certificates new_trust_anchor tl
                   | _ -> None
                     ) 
                   else None
                   
  | txt::tag :: [] -> ( if verify #certified trust_anchor txt tag 
                   (* then Some(txt, tag) *)
                   then Some (trust_anchor,
                    tag)
                   else None
                   )
  | _ -> None


val checkmail2: vkey certified -> data -> option (verified_mail)
(*TODO implement this function*)
(* The previous version only uses a single certificate.
 * This version should make use of a certificate chain.
 * See the figure below for an ilustration *)
let checkmail2 vk0 inbox = 
  (match parse inbox with 
  | [ text; cert; mtag ] ->
     (match check_certificates vk0 (parse cert) with
      | Some (ctxt, ctag) ->
      
        (match parse ctxt with
        | [ sender; vk ] ->
        if verify #(sent sender) vk text mtag then
               Some (ReceivedFrom sender text)
        else None
        | _ -> None)
       | None -> None )(* check_certificates doesn't output anything *)
  | _ -> None) (* parsing the inbox fails *)


(*
                            verify
            -------------------------------------------------------------------
            |                                                                 |
vk0         |   inbox                                                         |
 |          v     |                                                           |
 |        mail  mail_tag  certs                                               |
 |                          |                                                 |
 |                        cert1      ::      cert2      :: ... ::  certn      |
 |                          |                  |                     |        |
 |------------------->ctxt1   ctag1    ->ctxt2   ctag2      -->ctxtn   ctagn  |
    verify              |              |                    |    |---|        |
                       vk1-------------|            ...  ---|        |        |
                            verify                     verify   sender vkn ---|
*)

